"""Tests for ``highcharts.no_data``."""

import pytest

import datetime
from json.decoder import JSONDecodeError

from highcharts_core.options.series.data.boxplot import BoxPlotData as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'box_dash_style': 'Solid',
      'high': 654,
      'low': 321,
      'median': 456,
      'median_dash_style': 'Dash',
      'q1': 123,
      'q3': 567,
      'stem_dash_style': 'Solid',
      'whisker_dash_style': 'Solid'
    }, None),
    # Categorical X Value
    ({
      'box_dash_style': 'Solid',
      'high': 654,
      'low': 321,
      'median': 456,
      'median_dash_style': 'Dash',
      'q1': 123,
      'q3': 567,
      'stem_dash_style': 'Solid',
      'whisker_dash_style': 'Solid',

      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'drag_drop': {
          'draggableX': True,
          'draggableY': True,
          'dragHandle': {
              'className': 'draghandle-classname-goes-here',
              'color': '#ccc',
              'cursor': 'alias',
              'lineColor': '#ddd',
              'lineWidth': 2,
              'pathFormatter': """function() { return true; }""",
              'zIndex': 10
          },
          'dragMaxX': 3456,
          'dragMaxY': 6532,
          'dragMinX': 123,
          'dragMinY': 321,
          'dragPrecisionX': 5,
          'dragPrecisionY': 5,
          'dragSensitivity': 2,
          'groupBy': 'some-property-name',
          'guideBox': {
              'default': {
                  'className': 'some-classname-goes-here',
                  'color': '#999',
                  'cursor': 'pointer',
                  'lineColor': '#ccc',
                  'lineWidth': 2,
                  'zIndex': 100
              }
          },
          'liveRedraw': True
      },
      'drilldown': 'some-id-goes-here',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'x': 'some category',
      'y': 123
    }, None),
    # Datetime X Value
    ({
      'box_dash_style': 'Solid',
      'high': 654,
      'low': 321,
      'median': 456,
      'median_dash_style': 'Dash',
      'q1': 123,
      'q3': 567,
      'stem_dash_style': 'Solid',
      'whisker_dash_style': 'Solid',

      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'drag_drop': {
          'draggableX': True,
          'draggableY': True,
          'dragHandle': {
              'className': 'draghandle-classname-goes-here',
              'color': '#ccc',
              'cursor': 'alias',
              'lineColor': '#ddd',
              'lineWidth': 2,
              'pathFormatter': """function() { return true; }""",
              'zIndex': 10
          },
          'dragMaxX': 3456,
          'dragMaxY': 6532,
          'dragMinX': 123,
          'dragMinY': 321,
          'dragPrecisionX': 5,
          'dragPrecisionY': 5,
          'dragSensitivity': 2,
          'groupBy': 'some-property-name',
          'guideBox': {
              'default': {
                  'className': 'some-classname-goes-here',
                  'color': '#999',
                  'cursor': 'pointer',
                  'lineColor': '#ccc',
                  'lineWidth': 2,
                  'zIndex': 100
              }
          },
          'liveRedraw': True
      },
      'drilldown': 'some-id-goes-here',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'x': datetime.datetime(2022, 7, 26, 0, 4, 0),
      'y': 123
    }, None),
    # Date X Value
    ({
      'box_dash_style': 'Solid',
      'high': 654,
      'low': 321,
      'median': 456,
      'median_dash_style': 'Dash',
      'q1': 123,
      'q3': 567,
      'stem_dash_style': 'Solid',
      'whisker_dash_style': 'Solid',

      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'drag_drop': {
          'draggableX': True,
          'draggableY': True,
          'dragHandle': {
              'className': 'draghandle-classname-goes-here',
              'color': '#ccc',
              'cursor': 'alias',
              'lineColor': '#ddd',
              'lineWidth': 2,
              'pathFormatter': """function() { return true; }""",
              'zIndex': 10
          },
          'dragMaxX': 3456,
          'dragMaxY': 6532,
          'dragMinX': 123,
          'dragMinY': 321,
          'dragPrecisionX': 5,
          'dragPrecisionY': 5,
          'dragSensitivity': 2,
          'groupBy': 'some-property-name',
          'guideBox': {
              'default': {
                  'className': 'some-classname-goes-here',
                  'color': '#999',
                  'cursor': 'pointer',
                  'lineColor': '#ccc',
                  'lineWidth': 2,
                  'zIndex': 100
              }
          },
          'liveRedraw': True
      },
      'drilldown': 'some-id-goes-here',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'x': datetime.date(2022, 7, 26),
      'y': 123
    }, None),
    # + DataBase options
    # Categorical X Value
    ({
      'box_dash_style': 'Solid',
      'high': 654,
      'low': 321,
      'median': 456,
      'median_dash_style': 'Dash',
      'q1': 123,
      'q3': 567,
      'stem_dash_style': 'Solid',
      'whisker_dash_style': 'Solid',

      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'drag_drop': {
          'draggableX': True,
          'draggableY': True,
          'dragHandle': {
              'className': 'draghandle-classname-goes-here',
              'color': '#ccc',
              'cursor': 'alias',
              'lineColor': '#ddd',
              'lineWidth': 2,
              'pathFormatter': """function() { return true; }""",
              'zIndex': 10
          },
          'dragMaxX': 3456,
          'dragMaxY': 6532,
          'dragMinX': 123,
          'dragMinY': 321,
          'dragPrecisionX': 5,
          'dragPrecisionY': 5,
          'dragSensitivity': 2,
          'groupBy': 'some-property-name',
          'guideBox': {
              'default': {
                  'className': 'some-classname-goes-here',
                  'color': '#999',
                  'cursor': 'pointer',
                  'lineColor': '#ccc',
                  'lineWidth': 2,
                  'zIndex': 100
              }
          },
          'liveRedraw': True
      },
      'drilldown': 'some-id-goes-here',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'x': 'some category',
      'y': 123,

      'accessibility': {
          'description': 'Some description goes here',
          'enabled': True
      },
      'class_name': 'some-class-name',
      'color': '#ccc',
      'color_index': 2,
      'custom': {
          'some_key': 123,
          'other_key': 456
      },
      'description': 'Some description goes here',
      'events': {
        'click': """function(event) { return true; }""",
        'drag': """function(event) { return true; }""",
        'drop': """function(event) { return true; }""",
        'mouseOut': """function(event) { return true; }"""
      },
      'id': 'some-id-goes-here',
      'label_rank': 3,
      'name': 'Some Name Goes here',
      'selected': False
    }, None),
    # Datetime X Value
    ({
      'box_dash_style': 'Solid',
      'high': 654,
      'low': 321,
      'median': 456,
      'median_dash_style': 'Dash',
      'q1': 123,
      'q3': 567,
      'stem_dash_style': 'Solid',
      'whisker_dash_style': 'Solid',

      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'drag_drop': {
          'draggableX': True,
          'draggableY': True,
          'dragHandle': {
              'className': 'draghandle-classname-goes-here',
              'color': '#ccc',
              'cursor': 'alias',
              'lineColor': '#ddd',
              'lineWidth': 2,
              'pathFormatter': """function() { return true; }""",
              'zIndex': 10
          },
          'dragMaxX': 3456,
          'dragMaxY': 6532,
          'dragMinX': 123,
          'dragMinY': 321,
          'dragPrecisionX': 5,
          'dragPrecisionY': 5,
          'dragSensitivity': 2,
          'groupBy': 'some-property-name',
          'guideBox': {
              'default': {
                  'className': 'some-classname-goes-here',
                  'color': '#999',
                  'cursor': 'pointer',
                  'lineColor': '#ccc',
                  'lineWidth': 2,
                  'zIndex': 100
              }
          },
          'liveRedraw': True
      },
      'drilldown': 'some-id-goes-here',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'x': datetime.datetime(2022, 7, 26, 0, 4, 0),
      'y': 123,

      'accessibility': {
          'description': 'Some description goes here',
          'enabled': True
      },
      'class_name': 'some-class-name',
      'color': '#ccc',
      'color_index': 2,
      'custom': {
          'some_key': 123,
          'other_key': 456
      },
      'description': 'Some description goes here',
      'events': {
        'click': """function(event) { return true; }""",
        'drag': """function(event) { return true; }""",
        'drop': """function(event) { return true; }""",
        'mouseOut': """function(event) { return true; }"""
      },
      'id': 'some-id-goes-here',
      'label_rank': 3,
      'name': 'Some Name Goes here',
      'selected': False
    }, None),
    # Date X value
    ({
      'box_dash_style': 'Solid',
      'high': 654,
      'low': 321,
      'median': 456,
      'median_dash_style': 'Dash',
      'q1': 123,
      'q3': 567,
      'stem_dash_style': 'Solid',
      'whisker_dash_style': 'Solid',
      
      'data_labels': {
        'align': 'center',
        'allowOverlap': True,
        'animation': {
            'defer': 5
        },
        'backgroundColor': {
            'linearGradient': {
                'x1': 0.123,
                'x2': 0.234,
                'y1': 0.345,
                'y2': 0.456
            },
            'stops': [
                [0.12, '#999'],
                [0.34, '#fff']
            ]
        },
        'borderColor': '#999999',
        'borderRadius': 24,
        'borderWidth': 1,
        'className': 'some-class-name',
        'color': '#000000',
        'crop': True,
        'defer': False,
        'enabled': True,
        'filter': {
            'operator': '>=',
            'property': 'some_property',
            'value': 123
        },
        'format': 'some format',
        'formatter': """function() { return true; }""",
        'inside': True,
        'nullFormat': 'some format',
        'nullFormatter': """function() { return true; }""",
        'overflow': 'none',
        'padding': 12,
        'position': 'center',
        'rotation': 0,
        'shadow': False,
        'shape': 'rect',
        'style': 'style goes here',
        'useHTML': False,
        'verticalAlign': 'top',
        'x': 10,
        'y': 20,
        'z': 0
      },
      'drag_drop': {
          'draggableX': True,
          'draggableY': True,
          'dragHandle': {
              'className': 'draghandle-classname-goes-here',
              'color': '#ccc',
              'cursor': 'alias',
              'lineColor': '#ddd',
              'lineWidth': 2,
              'pathFormatter': """function() { return true; }""",
              'zIndex': 10
          },
          'dragMaxX': 3456,
          'dragMaxY': 6532,
          'dragMinX': 123,
          'dragMinY': 321,
          'dragPrecisionX': 5,
          'dragPrecisionY': 5,
          'dragSensitivity': 2,
          'groupBy': 'some-property-name',
          'guideBox': {
              'default': {
                  'className': 'some-classname-goes-here',
                  'color': '#999',
                  'cursor': 'pointer',
                  'lineColor': '#ccc',
                  'lineWidth': 2,
                  'zIndex': 100
              }
          },
          'liveRedraw': True
      },
      'drilldown': 'some-id-goes-here',
      'marker': {
        'enabled': True,
        'fillColor': '#cccccc',
        'height': 24,
        'lineWidth': 2,
        'radius': 2,
        'states': {
            'hover': {
                'enabled': True
            }
        },
        'symbol': 'circle',
        'width': 48
      },
      'x': datetime.date(2022, 7, 26),
      'y': 123,

      'accessibility': {
          'description': 'Some description goes here',
          'enabled': True
      },
      'class_name': 'some-class-name',
      'color': '#ccc',
      'color_index': 2,
      'custom': {
          'some_key': 123,
          'other_key': 456
      },
      'description': 'Some description goes here',
      'events': {
        'click': """function(event) { return true; }""",
        'drag': """function(event) { return true; }""",
        'drop': """function(event) { return true; }""",
        'mouseOut': """function(event) { return true; }"""
      },
      'id': 'some-id-goes-here',
      'label_rank': 3,
      'name': 'Some Name Goes here',
      'selected': False
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_BoxPlotData__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_BoxPlotData__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_BoxPlotData_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_BoxPlotData_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('series/data/boxplot/01.js', False, None),

    ('series/data/boxplot/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('series/data/boxplot/01.js', True, None),

    ('series/data/boxplot/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_BoxPlotData_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta


class ShadowOptions(HighchartsMeta):
    """Object configuring the shadow to apply to another object."""

    def __init__(self, **kwargs):
        self._color = None,
        self._offset_x = None
        self._offset_y = None
        self._opacity = None
        self._width = None

        self.color = kwargs.get('color', None)
        self.offset_x = kwargs.get('offset_x', None)
        self.offset_y = kwargs.get('offset_y', None)
        self.opacity = kwargs.get('opacity', None)
        self.width = kwargs.get('width', None)

    @property
    def color(self) -> Optional[str]:
        """The color to apply to the shadow.

        :returns: Color string
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = validators.string(value, allow_empty = True)

    @property
    def offset_x(self) -> Optional[int | float | Decimal]:
        """The offset to apply along the horizontal axis.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._offset_x

    @offset_x.setter
    def offset_x(self, value):
        self._offset_x = validators.numeric(value, allow_empty = True)

    @property
    def offset_y(self) -> Optional[int | float | Decimal]:
        """The offset to apply along the vertical axis.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._offset_y

    @offset_y.setter
    def offset_y(self, value):
        self._offset_y = validators.numeric(value, allow_empty = True)

    @property
    def opacity(self) -> Optional[int | float | Decimal]:
        """The opacity to apply to the shadow.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        self._opacity = validators.numeric(value, allow_empty = True)

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """The width of the shadow in pixels.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'color': as_dict.get('color', None),
            'offset_x': as_dict.get('offsetX', None),
            'offset_y': as_dict.get('offsetY', None),
            'opacity': as_dict.get('opacity', None),
            'width': as_dict.get('width', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'color': self.color,
            'offsetX': self.offset_x,
            'offsetY': self.offset_y,
            'opacity': self.opacity,
            'width': self.width
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.ast import AttributeObject
from highcharts_core.utility_classes.position import Position


class NoData(HighchartsMeta):
    """Options for displaying a message like "No data to display".

    .. warning::

      This feature requires the file ``no-data-to-display.js`` to be loaded in the
      page.

    .. tip::

      The actual text to display is set in the :meth:`language <Options.language>`
      options.

    """

    def __init__(self, **kwargs):
        self._attr = None
        self._position = None
        self._style = None
        self._use_html = None

        self.attr = kwargs.get('attr', None)
        self.position = kwargs.get('position', None)
        self.style = kwargs.get('style', None)
        self.use_html = kwargs.get('use_html', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'noData'

    @property
    def attr(self) -> Optional[AttributeObject]:
        """An object of additional SVG attributes for the no-data label.

        :rtype: :class:`AttributeObject` or :obj:`None <python:None>`
        """
        return self._attr

    @attr.setter
    @class_sensitive(AttributeObject)
    def attr(self, value):
        self._attr = value

    @property
    def position(self) -> Optional[Position]:
        """The position of the no-data label, relative to the plot area.

        :rtype: :class:`Position` or :obj:`None <python:None>`
        """
        return self._position

    @position.setter
    @class_sensitive(Position)
    def position(self, value):
        self._position = value

    @property
    def style(self) -> Optional[str]:
        """CSS styles to apply to the no-data label.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @property
    def use_html(self) -> Optional[bool]:
        """If ``True``, inserts the label as HTML. If ``False``, inserts the label as
        pseudo-HTML rendered with SVG. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>`
        """
        return self._use_html

    @use_html.setter
    def use_html(self, value):
        if value is None:
            self._use_html = None
        else:
            self._use_html = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'attr': as_dict.get('attr', None),
            'position': as_dict.get('position', None),
            'style': as_dict.get('style', None),
            'use_html': as_dict.get('useHTML', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'attr': self.attr,
            'position': self.position,
            'style': self.style,
            'useHTML': self.use_html
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta


class SonificationLanguageOptions(HighchartsMeta):
    """Language options for the sonification functionality when used in accessibility
    mode."""

    def __init__(self, **kwargs):
        self._play_as_sound_button_text = None
        self._play_as_sound_click_announcement = None

        self.play_as_sound_button_text = kwargs.get('play_as_sound_button_text', None)
        self.play_as_sound_click_announcement = kwargs.get(
            'play_as_sound_click_announcement',
            None
        )

    @property
    def play_as_sound_button_text(self) -> Optional[str]:
        """Defaults to
        ``'Play as sound, {chartTitle}'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._play_as_sound_button_text

    @play_as_sound_button_text.setter
    def play_as_sound_button_text(self, value):
        self._play_as_sound_button_text = validators.string(value, allow_empty = True)

    @property
    def play_as_sound_click_announcement(self) -> Optional[str]:
        """Defaults to
        ``'Play'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._play_as_sound_click_announcement

    @play_as_sound_click_announcement.setter
    def play_as_sound_click_announcement(self, value):
        self._play_as_sound_click_announcement = validators.string(value,
                                                                   allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'play_as_sound_button_text': as_dict.get('playAsSoundButtonText', None),
            'play_as_sound_click_announcement': as_dict.get(
                'playAsSoundClickAnnouncement',
                None
            ),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'playAsSoundButtonText': self.play_as_sound_button_text,
            'playAsSoundClickAnnouncement': self.play_as_sound_click_announcement
        }

        return untrimmed



"""Tests for ``highcharts.loading``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.loading import Loading as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
        'hide_duration': 100,
        'show_duration': 500
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('loading/01.js', False, None),
    ('loading/02.js', False, None),

    ('loading/error-01.js', False, (errors.HighchartsValueError,
                                    errors.HighchartsParseError,
                                    JSONDecodeError,
                                    TypeError,
                                    ValueError)),
    ('loading/error-02.js', False, (errors.HighchartsValueError,
                                    errors.HighchartsParseError,
                                    JSONDecodeError,
                                    ValueError)),

    ('loading/01.js', True, None),
    ('loading/02.js', True, None),

    ('loading/error-01.js', True, (errors.HighchartsValueError,
                                   errors.HighchartsParseError,
                                   JSONDecodeError,
                                   TypeError,
                                   ValueError)),
    ('loading/error-02.js', True, (errors.HighchartsValueError,
                                   errors.HighchartsParseError,
                                   JSONDecodeError,
                                   TypeError,
                                   ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from highcharts_core.utility_classes.animation import AnimationOptions
from highcharts_core.utility_classes.ast import (ASTMap,
                                                   ASTNode,
                                                   TextPath,
                                                   AttributeObject)
from highcharts_core.utility_classes.breadcrumbs import BreadcrumbOptions
from highcharts_core.utility_classes.buttons import (ExportingButtons,
                                                       ContextButtonConfiguration,
                                                       ButtonConfiguration,
                                                       ButtonTheme)
from highcharts_core.utility_classes.clusters import ClusterOptions
from highcharts_core.utility_classes.data_grouping import DataGroupingOptions
from highcharts_core.utility_classes.data_labels import DataLabel, NodeDataLabel
from highcharts_core.utility_classes.date_time_label_formats import DateTimeLabelFormats
from highcharts_core.utility_classes.events import *
from highcharts_core.utility_classes.gradients import (LinearGradient,
                                                         RadialGradient,
                                                         Gradient)
from highcharts_core.utility_classes.javascript_functions import (CallbackFunction,
                                                                    JavaScriptClass)
from highcharts_core.utility_classes.jitter import Jitter
from highcharts_core.utility_classes.markers import Marker
from highcharts_core.utility_classes.menus import MenuObject, MenuItem
from highcharts_core.utility_classes.nodes import (OrganizationNodeOptions,
                                                     DependencyWheelNodeOptions,
                                                     NodeOptions)
from highcharts_core.utility_classes.partial_fill import PartialFillOptions
from highcharts_core.utility_classes.patterns import Pattern, PatternOptions
from highcharts_core.utility_classes.position import Position
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.states import *
from highcharts_core.utility_classes.zones import ClusterZone, Zone



from highcharts_core.options.plot_options.dependencywheel import DependencyWheelOptions


class SankeyOptions(DependencyWheelOptions):
    """General options to apply to all Sankey series types.

    A sankey diagram is a type of flow diagram, in which the width of the link between
    two nodes is shown proportionally to the flow quantity.

    .. tabs::

      .. tab:: Standard Sankey

        .. figure:: ../../../_static/sankey-example.png
          :alt: Sankey Example Chart
          :align: center

      .. tab:: Inverted Sankey

        .. figure:: ../../../_static/sankey-example-inverted.png
          :alt: Inverted Sankey Example Chart
          :align: center

      .. tab:: Sankey with Outgoing Links

        .. figure:: ../../../_static/sankey-example-outgoing.png
          :alt: Sankey Example Chart with Outgoing Links
          :align: center

    """
    pass



"""Unit tests for ``highcharts.metaclasses``."""

import pytest

from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core import constants

from json.decoder import JSONDecodeError
from validator_collection import checkers


class TestClass(HighchartsMeta):
    """Class used to test the :class:`HighchartsMeta` functionality."""

    def __init__(self, **kwargs):
        self.item1 = kwargs.get('item1', None)
        self.item2 = kwargs.get('item2', None)

        super().__init__(**kwargs)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'item1': as_dict.get('item1', None),
            'item2': as_dict.get('item2', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'item1': self.item1,
            'item2': self.item2
        }


test_class_instance = TestClass(item1 = 123, item2 = 456)
test_class_trimmed_instance = TestClass(item1 = 123)
test_class_iterable = TestClass(item1 = [1, 2, constants.EnforcedNullType], item2 = 456)
test_class_none_iterable = TestClass(item1 = [1, None, 3], item2 = 456)

@pytest.mark.parametrize('kwargs, error', [
    ({'item1': 123,
      'item2': 456},
     None),
])
def test__init__(kwargs, error):
    if not error:
        result = TestClass(**kwargs)
        assert result is not None
        assert isinstance(result, HighchartsMeta)

    else:
        with pytest.raises(error):
            result = TestClass(**kwargs)


@pytest.mark.parametrize('cls, as_dict, error', [
    (TestClass, {'item1': 123, 'item2': 456}, None),
    (TestClass, {}, None),
    (TestClass, 'not-a-dict', (TypeError, AttributeError)),
])
def test_from_dict(cls, as_dict, error):
    if not error:
        original_dict = as_dict.copy()
        result = cls.from_dict(as_dict)
        assert result is not None
        assert isinstance(result, HighchartsMeta)
        print(original_dict)
        if not original_dict:
            assert result.item1 is None
            assert result.item2 is None
        else:
            for key in original_dict:
                attribute = getattr(result, key)
                assert attribute == original_dict[key]
    else:
        with pytest.raises(error):
            result = cls.from_dict(as_dict)


@pytest.mark.parametrize('cls, as_json, error', [
    (TestClass, '{"item1": 123, "item2": 456}', None),
    (TestClass, '{}', None),
    (TestClass, 'not-valid-json', JSONDecodeError),
])
def test_from_json(cls, as_json, error):
    if not error:
        result = cls.from_json(as_json)
        assert result is not None
        assert isinstance(result, HighchartsMeta)
    else:
        with pytest.raises(error):
            result = cls.from_json(as_json)


@pytest.mark.parametrize('instance, error', [
    (test_class_instance, None),
])
def test_to_json(instance, error):
    if not error:
        result = instance.to_json()
        assert result is not None
        assert isinstance(result, (str, bytes)) is True

        new_instance = instance.__class__.from_json(result)
        assert new_instance.item1 == instance.item1
        assert new_instance.item2 == instance.item2

    else:
        with pytest.raises(error):
            result = instance.to_json()


@pytest.mark.parametrize('instance, expected, error', [
    (test_class_instance, {'item1': 123, 'item2': 456}, None),
    (test_class_trimmed_instance, {'item1': 123}, None),
])
def test_to_dict(instance, expected, error):
    if not error:
        result = instance.to_dict()
        assert result is not None
        assert isinstance(result, dict) is True

        assert checkers.are_dicts_equivalent(result, expected) is True
    else:
        with pytest.raises(error):
            result = instance.to_dict()


@pytest.mark.parametrize('instance, error', [
    (test_class_trimmed_instance, None),
])
def test__to_untrimmed_dict(instance, error):
    if not error:
        result = instance._to_untrimmed_dict()
        assert result is not None
        assert isinstance(result, dict)
        assert len(result) == 2
        assert 'item1' in result
        assert 'item2' in result

        assert result.get('item1') is not None
        assert result.get('item2') is None
    else:
        with pytest.raises(error):
            result = instance._to_untrimmed_dict()


@pytest.mark.parametrize('untrimmed, expected_keys, error', [
    ({'item1': 123, 'item2': None}, 1, None),
    ({'item1': 123, 'item2': 456}, 2, None),
    ({'item1': test_class_instance, 'item2': None}, 1, None),
    ({'item1': constants.EnforcedNull, 'item2': None}, 1, None),
    ({'item1': {'test': 789}, 'item2': None}, 1, None),
    ('not-a-dict', None, AttributeError),
])
def test_trim_dict(untrimmed, expected_keys, error):
    if not error:
        result = HighchartsMeta.trim_dict(untrimmed)
        assert result is not None
        assert isinstance(result, dict) is True
        assert len(result) == expected_keys
    else:
        with pytest.raises(error):
            result = HighchartsMeta.trim_dict(untrimmed)


@pytest.mark.parametrize('untrimmed, expected_type, expected_length, error', [
    ([1, 2, 3], list, 3, None),
    (123, int, None, None),
    ([1, 2, None], list, 3, None),
    ([1, 2, constants.EnforcedNull], list, 3, None),
    ([1, 2, test_class_instance], list, 3, None),
])
def test_trim_iterable(untrimmed, expected_type, expected_length, error):
    if not error:
        result = TestClass.trim_iterable(untrimmed)
        assert result is not None
        assert isinstance(result, expected_type) is True
        if expected_length is not None:
            assert len(result) == expected_length
        if isinstance(untrimmed, list) and (None in untrimmed or
                                            constants.EnforcedNull in untrimmed):
            assert None not in result
            assert 'null' in result
    else:
        with pytest.raises(error):
            result = TestClass.trim_iterable(untrimmed)

"""
@pytest.mark.parametrize('error', [
    (None),
])
def test__mro_init(error):

    class ParentA(HighchartsMeta):
        def __init__(self, **kwargs):
            self._item1 = None
            self._item2 = None

            self.item1 = kwargs.get('item1', 123)
            self.item2 = kwargs.get('item2', 456)

        @property
        def item1(self):
            return self._item1

        @item1.setter
        def item1(self, value):
            self._item1 = value

        @property
        def item2(self):
            return self._item2

        @item2.setter
        def item2(self, value):
            self._item2 = value

        def _to_untrimmed_dict(self, in_cls = None) -> dict:
            return {
                'item1': self.item1,
                'item2': self.item2
            }

        @classmethod
        def from_dict(cls, as_dict):
            return cls(**as_dict)

    class ParentB(HighchartsMeta):
        def __init__(self, **kwargs):
            self._item3 = None
            self._item4 = None

            self.item3 = kwargs.get('item3', 789)
            self.item4 = kwargs.get('item4', 987)

        @property
        def item3(self):
            return self._item3

        @item3.setter
        def item3(self, value):
            self._item3 = value

        @property
        def item4(self):
            return self._item4

        @item4.setter
        def item4(self, value):
            self._item4 = value

        def _to_untrimmed_dict(self, in_cls = None) -> dict:
            return {
                'item3': self.item3,
                'item4': self.item4
            }

        @classmethod
        def from_dict(cls, as_dict):
            return cls(**as_dict)

    class Child(ParentA, ParentB):
        def __init__(self, **kwargs):
            self._item1 = None
            self._item2 = None
            self._item3 = None
            self._item4 = None

            self.child_item = kwargs.get('child_item', 'test value')

        def to_dict(self) -> dict:
            return {
                'child_item': self.child_item
            }

    class ChildError(ParentA):
        pass

    if not error:
        child_instance = Child()

        assert child_instance.item1 is None
        assert child_instance.item2 is None
        assert child_instance.item3 is None
        assert child_instance.item4 is None

        kwargs = {}
        child_instance.__mro_init__(kwargs)

        assert child_instance.item1 is not None
        assert child_instance.item2 is not None
        assert child_instance.item3 is not None
        assert child_instance.item4 is not None

    else:
        child_instance = ChildError()

        with pytest.raises(error):
            child_instance.__mro_init__({})
"""

@pytest.mark.parametrize('cls, as_str, error', [
    (TestClass, """{item1: function () {
      // The first returned item is the header, subsequent items are the
      // points
      return ['<b>' + this.x + '</b>'].concat(
          this.points ?
              this.points.map(function (point) {
                  return point.series.name + ': ' + point.y + 'm';
              }) : []
      );
     }}""", None)
])
def test_from_js_literal(cls, as_str, error):
    if not error:
        print('-------------------')
        print('ORIGINAL VALIDATION')
        parsed_original, original_str = cls._validate_js_literal(as_str, range = False)
        print('-------------')
        print('ORIGINAL CALL')
        result = cls.from_js_literal(as_str)
        assert result is not None
        assert isinstance(result, cls) is True

        as_js_literal = result.to_js_literal()
        print('-----------------')
        print('RESULT VALIDATION')
        parsed_output, output_str = cls._validate_js_literal(as_js_literal, range = False)
        assert str(parsed_output) == str(parsed_original)
    else:
        with pytest.raises(error):
            result = cls.from_js_literal(as_str)



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import errors, constants, utility_functions
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core.utility_classes.animation import AnimationOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.events import ChartEvents
from highcharts_core.options.chart.options_3d import Options3D
from highcharts_core.options.axes.parallel_axes import ParallelAxesOptions
from highcharts_core.options.chart.scrollable_plot_area import ScrollablePlotArea
from highcharts_core.options.chart.zooming import ZoomingOptions


class PanningOptions(HighchartsMeta):
    """Configures panning behavior in a chart.

    .. hint::

      Best used with :meth:`Chart.pan_key` to combine zooming and panning.

    .. note::

      On touch devices, when the :meth:`Tooltip.follow_touch_move` property is
      ``True`` (default), panning requires two fingers. To allow panning with one
      finger, set ``follow_touch_move`` to ``False``.

    """

    def __init__(self, **kwargs):
        self._enabled = None
        self._type = None

        self.enabled = kwargs.get('enabled', None)
        self.type = kwargs.get('type', None)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables chart panning. Defaults to ``False``.

        :returns: Flag enabling or disabling chart panning.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def type(self) -> Optional[str]:
        """Determines in what dimensions the user can pan the chart. Defaults to ``'x'``.

        Accepts:

          * ``'x'``
          * ``'y'``
          * ``'xy'``

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._type

    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['x', 'y', 'xy']:
                raise errors.HighchartsValueError(f'Panning.type expects a value of '
                                                  f'either "x", "y", "xy", or None. '
                                                  f'Received: {value}')
            self._type = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None),
            'type': as_dict.get('type', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'enabled': self.enabled,
            'type': self.type
        }


class ChartOptions(HighchartsMeta):
    """Configuration settings that apply to a chart."""

    def __init__(self, **kwargs):
        self._align_thresholds = None
        self._align_ticks = None
        self._allow_mutating_data = None
        self._animation = None
        self._background_color = None
        self._border_color = None
        self._border_radius = None
        self._border_width = None
        self._class_name = None
        self._color_count = None
        self._display_errors = None
        self._events = None
        self._height = None
        self._ignore_hidden_series = None
        self._inverted = None
        self._margin_bottom = None
        self._margin_left = None
        self._margin_right = None
        self._margin_top = None
        self._number_formatter = None
        self._options_3d = None
        self._pan_key = None
        self._panning = None
        self._parallel_axes = None
        self._parallel_coordinates = None
        self._plot_background_color = None
        self._plot_background_image = None
        self._plot_border_color = None
        self._plot_border_width = None
        self._plot_shadow = None
        self._polar = None
        self._reflow = None
        self._render_to = None
        self._scrollable_plot_area = None
        self._selection_marker_fill = None
        self._shadow = None
        self._show_axes = None
        self._spacing_bottom = None
        self._spacing_left = None
        self._spacing_top = None
        self._spacing_right = None
        self._style = None
        self._styled_mode = None
        self._type = None
        self._width = None
        self._zooming = None

        self.align_thresholds = kwargs.get('align_thresholds', None)
        self.align_ticks = kwargs.get('align_ticks', None)
        self.allow_mutating_data = kwargs.get('allow_mutating_data', None)
        self.animation = kwargs.get('animation', None)
        self.background_color = kwargs.get('background_color', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.border_width = kwargs.get('border_width', None)
        self.class_name = kwargs.get('class_name', None)
        self.color_count = kwargs.get('color_count', None)
        self.display_errors = kwargs.get('display_errors', None)
        self.events = kwargs.get('events', None)
        self.height = kwargs.get('height', None)
        self.ignore_hidden_series = kwargs.get('ignore_hidden_series', None)
        self.inverted = kwargs.get('inverted', None)
        self.margin = kwargs.get('margin', None)
        if not kwargs.get('margin', None):
            self.margin_bottom = kwargs.get('margin_bottom', None)
            self.margin_left = kwargs.get('margin_left', None)
            self.margin_right = kwargs.get('margin_right', None)
            self.margin_top = kwargs.get('margin_top', None)
        self.number_formatter = kwargs.get('number_formatter', None)
        self.options_3d = kwargs.get('options_3d', None)
        self.pan_key = kwargs.get('pan_key', None)
        self.panning = kwargs.get('panning', None)
        self.parallel_axes = kwargs.get('parallel_axes', None)
        self.parallel_coordinates = kwargs.get('parallel_coordinates', None)
        self.plot_background_color = kwargs.get('plot_background_color', None)
        self.plot_background_image = kwargs.get('plot_background_image', None)
        self.plot_border_color = kwargs.get('plot_border_color', None)
        self.plot_border_width = kwargs.get('plot_border_width', None)
        self.plot_shadow = kwargs.get('plot_shadow', None)
        self.polar = kwargs.get('polar', None)
        self.reflow = kwargs.get('reflow', None)
        self.render_to = kwargs.get('render_to', None)
        self.scrollable_plot_area = kwargs.get('scrollable_plot_area', None)
        self.selection_marker_fill = kwargs.get('selection_marker_fill', None)
        self.shadow = kwargs.get('shadow', None)
        self.show_axes = kwargs.get('show_axes', None)
        self.spacing = kwargs.get('spacing', None)
        if not kwargs.get('spacing', None):
            self.spacing_bottom = kwargs.get('spacing_bottom', None)
            self.spacing_left = kwargs.get('spacing_left', None)
            self.spacing_top = kwargs.get('spacing_top', None)
            self.spacing_right = kwargs.get('spacing_right', None)
        self.style = kwargs.get('style', None)
        self.styled_mode = kwargs.get('styled_mode', None)
        self.type = kwargs.get('type', None)
        self.width = kwargs.get('width', None)
        self.zooming = kwargs.get('zooming', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'chart'

    @property
    def align_thresholds(self) -> Optional[bool]:
        """When using multiple axes, align the thresholds. When ``True``, other ticks will
        also be aligned. Defaults to ``False``.

        .. note::

          For line series and some other series types, the threshold option is set to
          ``null`` by default. This will in turn cause their y-axis to not have a
          threshold. In order to avoid that, set the series threshold to ``0`` or another
          number.

        .. note::

          If :meth:`AxisOptions.start_on_tick` or :meth:`AxisOptions.end_on_tick` are set
          to ``False``, or if the axis is logarithmic, the threshold will not be aligned.

        :returns: Flag indicating whether thresholds should be aligned.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._align_thresholds

    @align_thresholds.setter
    def align_thresholds(self, value):
        if value is None:
            self._align_thresholds = None
        else:
            self._align_thresholds = bool(value)

    @property
    def align_ticks(self) -> Optional[bool]:
        """If ``True`` and using multiple axes, the ticks of two or more opposite axes
        will automatically be aligned by adding ticks to the axis or axes with the least
        ticks, as if ``tick_amount`` were specified. This can be prevented by setting
        ``align_ticks`` to ``False``.

        Defaults to ``True``.

        .. hint::

          If the grid lines look messy, it's a good idea to hide them for the secondary
          axis by setting ``grid_line_width`` to ``0``.

        .. warning::

          If ``start_on_tick`` or ``end_on_tick`` in the axis options are set to
          ``False``, then tick alignment will be disabled for the axis.

        .. warning::

          Aways disabled for logarithmic axes.

        :returns: Flag indicating whether ot align ticks along the axes.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._align_ticks

    @align_ticks.setter
    def align_ticks(self, value):
        if value is None:
            self._align_ticks = None
        else:
            self._align_ticks = bool(value)

    @property
    def allow_mutating_data(self) -> Optional[bool]:
        """If ``True``, the original data source will be allowed to be mutated. However,
        if ``False`` then Highcharts will create a copy of the original data to prevent
        inadvertent mutation. Defaults to ``True``.

        .. note::

          By default, (for memory and performance reasons) the chart does not copy the
          data but keeps it as a reference.

          In some cases, this might result in mutating the original data source. In order
          to prevent that, set this property to ``False``. Please note that changing this
          may decrease performance, especially with bigger sets of data.

        :returns: Flag indicating whether to allow mutation by referencing th eoriginal
          data (``True``), or prevent it by copying the original data (``False``).
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_mutating_data

    @allow_mutating_data.setter
    def allow_mutating_data(self, value):
        if value is None:
            self._allow_mutating_data = None
        else:
            self._allow_mutating_data = bool(value)

    @property
    def animation(self) -> Optional[bool | AnimationOptions]:
        """Configures the overall animation for all chart updating.

        The animation can be configured as either a boolean or a :class:`AnimationOptions`
        object. If ``True``, it will apply the ``'swing'`` jQuery easing and a duration of
        500 ms by default. If used as a :class:`AnimationOptions` instance, you have more
        fine-grianed configuration control.

        Animation can be disabled throughout the chart by setting it to ``False`` here.
        It can be overridden for each individual API method as a function parameter. The
        only animation not affected by this option is the initial series animation, see
        :meth:`PlotOptions.series.animation`.

        When zooming on a series with less than 100 points, the chart redraw will be done
        with animation, but in case of more data points, it is necessary to set this
        option to ensure animation on zoom.

        :returns: Settings for the animation of image patterns.
        :rtype: :class:`AnimationOptions` or :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    def animation(self, value):
        if value is None:
            self._animation = None
        elif value is False:
            self._animation = False
        elif value is True:
            self._animation = True
        else:
            self._animation = validate_types(value,
                                             types = AnimationOptions)

    @property
    def background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or gradient for the outer chart area. Defaults to
        ``'#ffffff'``.

        :returns: The background color for the outer chart area.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        from highcharts_core import utility_functions
        self._background_color = utility_functions.validate_color(value)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the outer chart border. Defaults to
        ``'#335cad'``.

        :returns: The color of the outer chart border.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The border radius (in pixels) applied to the outer chart border. Defaults to
        ``0``.

        :returns: The border radius to apply to the outer chart border.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value, allow_empty = True)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The border width (in pixels) applied to the outer chart border. Defaults to
        ``0``.

        :returns: The border width to apply to the outer chart border.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value, allow_empty = True)

    @property
    def class_name(self) -> Optional[str]:
        """A classname to apply styling using CSS.

        :returns: The classname to apply to enable styling via CSS.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color_count(self) -> Optional[int]:
        """In styled mode, sets how many colors the class names should rotate between.
        Defaults to ``10``

        With ten colors, series (or points) are given class names like
        ``highcharts-color-0``, ``highcharts-color-0`` [...] ``highcharts-color-9``. The
        equivalent in non-styled mode is to set colors using the colors setting.

        :returns: The number of colors to cycle through in styled mode.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_count

    @color_count.setter
    def color_count(self, value):
        self._color_count = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0,
                                               coerce_value = True)

    @property
    def display_errors(self) -> Optional[bool]:
        """If ``True``, will display errors on the chart itself. If ``False``, will only
        report errors to the console. Defaults to ``True``.

        :returns: Flag indicating whether ot display errors on the chart.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._display_errors

    @display_errors.setter
    def display_errors(self, value):
        if value is None:
            self._display_errors = None
        else:
            self._display_errors = bool(value)

    @property
    def events(self) -> Optional[ChartEvents]:
        """Definition of JavaScript event listeners to apply to the chart.

        :rtype: :class:`ChartEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(ChartEvents)
    def events(self, value):
        self._events = value

    @property
    def height(self) -> Optional[constants.EnforcedNullType | int | float | Decimal | str]:
        """An explicit height for the chart. Defaults to :obj:`None <python:None>`.

        If a number, the height is given in pixels. If given a percentage string (for
        example ``'56%'``), the height is given as the percentage of the actual chart
        width. This allows for preserving the aspect ratio across responsive sizes.

        By default (when :obj:`None python:None>`) the height is calculated from the
        offset height of the containing element, or 400 pixels if the containing element's
        height is 0.

        :rtype: :class:`EnforcedNullType` or numeric or :class:`str <python:str>` or
          :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        if value is None or isinstance(value, constants.EnforcedNullType):
            self._height = None
        else:
            try:
                try:
                    self._height = validators.numeric(value,
                                                      allow_empty = False,
                                                      minimum = 0)
                except ValueError:
                    self._height = validators.string(value,
                                                     allow_empty = False)
            except ValueError:
                raise errors.HighchartsValueError('Unable to resolve value to a '
                                                  'supported data type.')

    @property
    def ignore_hidden_series(self) -> Optional[bool]:
        """If ``True``, the axes will scale to the remaining visible series once one
        series is hidden. If ``False``, hiding and showing a series will not affect the
        axes or the other series.

        Defaults to ``True``.

        .. note::

          For stacks, once one series within the stack is hidden, the rest of the stack
          will close in around it even if the axis is not affected.

        :returns: Flag indicating whether to ignore hidden series.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._ignore_hidden_series

    @ignore_hidden_series.setter
    def ignore_hidden_series(self, value):
        if value is None:
            self._ignore_hidden_series = None
        else:
            self._ignore_hidden_series = bool(value)

    @property
    def inverted(self) -> Optional[bool]:
        """If ``True``, inverts the axes so that the x-axis is vertical and y-axis is
        horizontal. Defaults to ``False``.

        .. hint::

          When ``True``, the x-axis is reversed by default.

        .. warning::

          If a bar series is present in the chart, it will be inverted automatically.

        .. hint::

          Inverting the chart doesn't have an effect if there are no cartesian series in
          the chart, or if the chart is a polar chart type.

        :returns: Flag indicating whether to invert the axes.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._inverted

    @inverted.setter
    def inverted(self, value):
        if value is None:
            self._inverted = None
        else:
            self._inverted = bool(value)

    @property
    def margin(self) -> Optional[List[int | float | Decimal]]:
        """The margin between the outer edge of the chart and the plot area. The numbers
        in the array designate top, right, bottom and left respectively.

        By default there is no margin. The actual space is dynamically calculated from the
        offset of axis labels, axis title, title, subtitle and legend in addition to the
        ``spacing_top``, ``spacing_right``, ``spacing_bottom`` and ``spacing_left``
        options.

        .. note::

          This property is a convenience property to consolidate simple margin
          configuration across the ``margin_top``, ``margin_right``, ``margin_bottom``,
          and ``margin_left`` options. Values will be propagated between all options.

        :rtype: :class:`list <python:list>` of four numeric values or
          :obj:`None <python:None>`
        """
        result = [self.margin_top,
                  self.margin_right,
                  self.margin_bottom,
                  self.margin_left]

        if not any(result):
            return None

        return result

    @margin.setter
    def margin(self, value):
        if value is None:
            self.margin_top = None
            self.margin_right = None
            self.margin_bottom = None
            self.margin_left = None
        elif checkers.is_iterable(value):
            if len(value) != 4:
                raise errors.HighchartsValueError(f'margin expects either a single value '
                                                  f'or an iterable of four values. '
                                                  f'Received an iterable of {len(value)} '
                                                  f'values ({value})')
            self.margin_top = value[0]
            self.margin_right = value[1]
            self.margin_bottom = value[2]
            self.margin_left = value[3]
        else:
            self.margin_top = value
            self.margin_right = value
            self.margin_bottom = value
            self.margin_left = value

    @property
    def margin_bottom(self) -> Optional[int | float | Decimal]:
        """The margin between the bottom outer edge of the chart and the plot area. Use
        this to set a fixed pixel value for the margin as opposed to the default dynamic
        margin.

        By default, not set.

        .. seealso::

          * :meth:`Chart.spacing_bottom`

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin_bottom

    @margin_bottom.setter
    def margin_bottom(self, value):
        self._margin_bottom = validators.numeric(value, allow_empty = True)

    @property
    def margin_left(self) -> Optional[int | float | Decimal]:
        """The margin between the left outer edge of the chart and the plot area. Use
        this to set a fixed pixel value for the margin as opposed to the default dynamic
        margin.

        By default, not set.

        .. seealso::

          * :meth:`Chart.spacing_left`

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin_left

    @margin_left.setter
    def margin_left(self, value):
        self._margin_left = validators.numeric(value, allow_empty = True)

    @property
    def margin_right(self) -> Optional[int | float | Decimal]:
        """The margin between the right outer edge of the chart and the plot area. Use
        this to set a fixed pixel value for the margin as opposed to the default dynamic
        margin.

        By default, not set.

        .. seealso::

          * :meth:`Chart.spacing_right`

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin_right

    @margin_right.setter
    def margin_right(self, value):
        self._margin_right = validators.numeric(value, allow_empty = True)

    @property
    def margin_top(self) -> Optional[int | float | Decimal]:
        """The margin between the top outer edge of the chart and the plot area. Use
        this to set a fixed pixel value for the margin as opposed to the default dynamic
        margin.

        By default, not set.

        .. seealso::

          * :meth:`Chart.spacing_top`

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin_top

    @margin_top.setter
    def margin_top(self, value):
        self._margin_top = validators.numeric(value, allow_empty = True)

    @property
    def number_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript Callback function to override the default function that formats all
        the numbers in the chart. Returns a string with the formatted number.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._number_formatter

    @number_formatter.setter
    @class_sensitive(CallbackFunction)
    def number_formatter(self, value):
        self._number_formatter = value

    @property
    def options_3d(self) -> Optional[Options3D]:
        """Options to render charts in three dimensions.

        .. note::

          This feature requires the JavaScript ``highcharts-3d.js`` module, found in the
          download package or online at
          `code.highcharts.com/highcharts-3d.js <https://code.highcharts.com/highcharts-3d.js>`_.

        :rtype: :class:`Options3D` or :obj:`None <python:None>`
        """
        return self._options_3d

    @options_3d.setter
    @class_sensitive(Options3D)
    def options_3d(self, value):
        self._options_3d = value

    @property
    def pan_key(self) -> Optional[str]:
        """Allows setting a key to switch between zooming and panning.

        Accepts the following values:

          * ``'alt'``
          * ``'ctrl'``
          * ``'meta'`` (the command key on Mac and Windows key on Windows)
          * ``'shift'``.

        The keys are mapped directly to the key properties of the click event argument
        (``event.altKey``, ``event.ctrlKey``, ``event.metaKey``, and ``event.shiftKey``).

        :returns: The key that switches between zooming and panning.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pan_key

    @pan_key.setter
    def pan_key(self, value):
        if not value:
            self._pan_key = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['alt', 'ctrl', 'meta', 'shift']:
                raise errors.HighchartsValueError(f'pan_key expects a value of "alt", '
                                                  f'"ctrl", "meta", or "shift". Was: '
                                                  f'{value}')
            self._pan_key = value

    @property
    def panning(self) -> Optional[PanningOptions]:
        """Configures panning behavior in a chart.

        .. hint::

          Best used with :meth:`Chart.pan_key` to combine zooming and panning.

        .. note::

          On touch devices, when the :meth:`Tooltip.follow_touch_move` property is
          ``True`` (default), panning requires two fingers. To allow panning with one
          finger, set ``follow_touch_move`` to ``False``.

        :returns: Configuration settings for panning behavior.
        :rtype: :class:`PanningOptions` or :obj:`None <python:None>`
        """
        return self._panning

    @panning.setter
    def panning(self, value):
        if value is None:
            self._panning = None
        else:
            if isinstance(value, bool):
                value = {
                    'enabled': value
                }
            value = validate_types(value, types = PanningOptions)
            
        self._panning = value

    @property
    def parallel_axes(self) -> Optional[ParallelAxesOptions]:
        """Common options for all Y-Axes rendered in a parallel coordinates plot.

        .. note::

          This feature requires in JavaScript ``modules/parallel-coordinates.js``.

        :rtype: :class:`ParallelAxesOptions` or :obj:`None <python:None>`
        """
        return self._parallel_axes

    @parallel_axes.setter
    @class_sensitive(ParallelAxesOptions)
    def parallel_axes(self, value):
        self._parallel_axes = value

    @property
    def parallel_coordinates(self) -> Optional[bool]:
        """If ``True``, renders charts as a parallel coordinates plot. Defaults to
        ``False``.

        In a parallel coordinates plot (||-coords) by default all required Y-axes are
        generated and the legend is disabled.

        .. note::

          This feature requires the JavaScript module ``modules/parallel-coordinates.js``.

        .. seealso::

          * `Parallel Coordinates Demo <https://jsfiddle.net/gh/get/library/pure/highcharts/highcharts/tree/master/samples//highcharts/demo/parallel-coordinates/>`_

        :returns: Flag indicating whether the chart is a parallel coordinates plot.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._parallel_coordinates

    @parallel_coordinates.setter
    def parallel_coordinates(self, value):
        if value is None:
            self._parallel_coordinates = None
        else:
            self._parallel_coordinates = bool(value)

    @property
    def plot_background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or gradient for the plot area. Defaults to
        ``None``.

        :returns: The background color for the plot area.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._plot_background_color

    @plot_background_color.setter
    def plot_background_color(self, value):
        from highcharts_core import utility_functions
        self._plot_background_color = utility_functions.validate_color(value)

    @property
    def plot_background_image(self) -> Optional[str]:
        """The URL for an image to use as the background of the plot area.

        .. hint::

           To set an image as the background for the entire chart, set a CSS background
           image to the container element.

         .. note::

           Note that for the image to be applied to exported charts, its URL needs to be
           accessible by the export server.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises HighchartsValueError: if a value is supplied that is not a URL or not
          path-like

        """
        return self._plot_background_image

    @plot_background_image.setter
    def plot_background_image(self, value):
        if not value:
            self._plot_background_image = None
        else:
            try:
                self._plot_background_image = validators.url(value)
            except ValueError:
                try:
                    self._plot_background_image = validators.path(value)
                except ValueError:
                    raise errors.HighchartsValueError(f'value provided ({value}) not a '
                                                      f'valid URL or path')

    @property
    def plot_border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the outer chart border. Defaults to
        ``'#cccccc'``.

        :returns: The color of the outer chart border.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._plot_border_color

    @plot_border_color.setter
    def plot_border_color(self, value):
        from highcharts_core import utility_functions
        self._plot_border_color = utility_functions.validate_color(value)

    @property
    def plot_border_width(self) -> Optional[int | float | Decimal]:
        """The border width (in pixels) applied to the outer chart border. Defaults to
        ``0``.

        :returns: The border width to apply to the outer chart border.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._plot_border_width

    @plot_border_width.setter
    def plot_border_width(self, value):
        self._plot_border_width = validators.numeric(value, allow_empty = True)

    @property
    def plot_shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration of a drop shadow applied to the plot area. Accepts either a
        boolean value of ``False`` which disables any shadow, or a :class:`ShadowOptions`
        instance with the applicable configuration.

        Defaults to ``False``.

        .. warning::

          Requires that :meth:`Chart.plot_background_color` be set.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._plot_shadow

    @plot_shadow.setter
    def plot_shadow(self, value):
        if value is None:
            self._plot_shadow = None
        elif value is False:
            self._plot_shadow = False
        else:
            value = validate_types(value,
                                   types = ShadowOptions,
                                   allow_none = False)
            self._plot_shadow = value

    @property
    def polar(self) -> Optional[bool]:
        """If ``True``, cartesian charts like line, spline, area, and column are
        transformed into the polar coordinate system. This produces polar charts (also
        known as radar charts). Defaults to ``False``.

        :returns: Flag indicating whether to render the chart as a polar chart.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._polar

    @polar.setter
    def polar(self, value):
        if value is None:
            self._polar = None
        else:
            self._polar = bool(value)

    @property
    def reflow(self) -> Optional[bool]:
        """If ``True``, reflows the chart to fit the width of the container ``div`` when
        the window is resized. Defaults to ``True``.

        :returns: Flag indicating whether to reflow the chart in response to window
          resizing.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._reflow

    @reflow.setter
    def reflow(self, value):
        if value is None:
            self._reflow = None
        else:
            self._reflow = bool(value)

    @property
    def render_to(self) -> Optional[str]:
        """The ID of the HTML element where the chart will be rendered.

        .. note::

          In JavaScript, the HTML element can also be passed by direct reference, or as
          the first argument of the chart constructor, in which case the option is not
          needed.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._render_to

    @render_to.setter
    def render_to(self, value):
        self._render_to = validators.string(value, allow_empty = True)

    @property
    def scrollable_plot_area(self) -> Optional[ScrollablePlotArea]:
        """Configuration settings to make the plot area scrollable.

        This feature provides a minimum size for the plot area of the chart. If the size
        gets smaller than this, typically on mobile devices, a native browser scrollbar is
        presented. This scrollbar provides smooth scrolling for the contents of the plot
        area, whereas the title, legend and unaffected axes are fixed.

        .. hint::

          Since v7.1.2, a scrollable plot area can be defined for either horizontal or
          vertical scrolling, depending on whether the `minimum_width` or `minimum_height`
          options are set.

        :rtype: :class:`ScrollablePlotArea` or :obj:`None <python:None>`
        """
        return self._scrollable_plot_area

    @scrollable_plot_area.setter
    @class_sensitive(ScrollablePlotArea)
    def scrollable_plot_area(self, value):
        self._scrollable_plot_area = value

    @property
    def selection_marker_fill(self) -> Optional[str | Gradient | Pattern]:
        """The background color or the marker square when selecting (zooming in on) an
        area of the chart. Defaults to
        ``'rgba(51,92,173,0.25)'``.

        :returns: The backgorund color of the marker square when selecting an area of the
          chart.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._selection_marker_fill

    @selection_marker_fill.setter
    def selection_marker_fill(self, value):
        self._selection_marker_fill = utility_functions.validate_color(value)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration of a drop shadow applied to the outer chart area. Accepts either
        a boolean value of ``False`` which disables any shadow, or a
        :class:`ShadowOptions` instance with the applicable configuration.

        Defaults to ``False``.

        .. warning::

          Requires that :meth:`Chart.background_color` be set.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if value is None:
            self._shadow = None
        elif value is False:
            self._shadow = False
        else:
            value = validate_types(value,
                                   types = ShadowOptions,
                                   allow_none = False)
            self._shadow = value

    @property
    def show_axes(self) -> Optional[bool]:
        """If ``True``, shows axes initially (before series have been added to the chart).

        .. warning::

          This property only applies to empty charts where series are added dynamically.

        :returns: Flag indicating whether to show axes on empty dynamic charts before
          series have been added.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_axes

    @show_axes.setter
    def show_axes(self, value):
        if value is None:
            self._show_axes = None
        else:
            self._show_axes = bool(value)

    @property
    def spacing(self) -> Optional[List[int | float | Decimal]]:
        """The distance between the outer edge of the chart and the content, like title
        or legend, or axis title and labels if present. The numbers in the array designate
        top, right, bottom and left respectively.

        By default, ``[10,10,15,10]``.

        .. note::

          This property is a convenience property to consolidate simple spacing
          configuration across the ``spacing_top``, ``spacing_right``, ``spacing_bottom``,
          and ``spacing_left`` options. Values will be propagated between all options.

        :rtype: :class:`list <python:list>` of four numeric values or
          :obj:`None <python:None>`
        """
        result = [self.spacing_top,
                  self.spacing_right,
                  self.spacing_bottom,
                  self.spacing_left]

        if not any(result):
            return None

        return result

    @spacing.setter
    def spacing(self, value):
        if value is None:
            self.spacing_top = None
            self.spacing_right = None
            self.spacing_bottom = None
            self.spacing_left = None
        elif checkers.is_iterable(value):
            if len(value) != 4:
                raise errors.HighchartsValueError(f'spacing expects either a single value'
                                                  f' or an iterable of four values. '
                                                  f'Received an iterable of {len(value)} '
                                                  f'values ({value})')
            value = [validators.numeric(x) for x in value]
            self.spacing_top = value[0]
            self.spacing_right = value[1]
            self.spacing_bottom = value[2]
            self.spacing_left = value[3]
        else:
            value = validators.numeric(value, allow_empty = False)
            self.spacing_top = value
            self.spacing_right = value
            self.spacing_bottom = value
            self.spacing_left = value

    @property
    def spacing_bottom(self) -> Optional[int | float | Decimal]:
        """The spacing between the bottom edge of the chart and the content (plot area,
        axis title and labels, title, subtitle or legend in top position).

        By default, ``15``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._spacing_bottom

    @spacing_bottom.setter
    def spacing_bottom(self, value):
        self._spacing_bottom = validators.numeric(value, allow_empty = True)

    @property
    def spacing_left(self) -> Optional[int | float | Decimal]:
        """The spacing between the left edge of the chart and the content (plot area,
        axis title and labels, title, subtitle or legend in top position).

        Defaults to ``10``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._spacing_left

    @spacing_left.setter
    def spacing_left(self, value):
        self._spacing_left = validators.numeric(value, allow_empty = True)

    @property
    def spacing_right(self) -> Optional[int | float | Decimal]:
        """The spacing between the right edge of the chart and the content (plot area,
        axis title and labels, title, subtitle or legend in top position).

        Defaults to ``10``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._spacing_right

    @spacing_right.setter
    def spacing_right(self, value):
        self._spacing_right = validators.numeric(value, allow_empty = True)

    @property
    def spacing_top(self) -> Optional[int | float | Decimal]:
        """The spacing between the top edge of the chart and the content (plot area, axis
        title and labels, title, subtitle or legend in top position).

        Defaults to ``10``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._spacing_top

    @spacing_top.setter
    def spacing_top(self, value):
        self._spacing_top = validators.numeric(value, allow_empty = True)

    @property
    def style(self) -> Optional[str]:
        """Additional CSS styles to apply inline to the container div.

        Defaults to ``'{"fontFamily": "\"Lucida Grande\", \"Lucida Sans Unicode\", Verdana, Arial, Helvetica, sans-serif","fontSize":"12px"}'``.

        .. note::

          Since the default font styles are applied in the renderer, it is ignorant of the
          individual chart options and must be set globally.

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @property
    def styled_mode(self) -> Optional[bool]:
        """If ``True``, sets the chart to operate in **styled mode**, where no
        presentational attributes or CSS are applied to the chart SVG. Instead, CSS rules
        are required to style the chart. Defaults to ``False``.

        .. seealso::

          * `Available CSS Styles and Variables <https://www.highcharts.com/docs/chart-design-and-style/style-by-css>`__.
          * The Default Style Sheet:
            `https://code.highcharts.com/css/highcharts.css <https://code.highcharts.com/css/highcharts.css>`_.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._styled_mode

    @styled_mode.setter
    def styled_mode(self, value):
        if value is None:
            self._styled_mode = None
        else:
            self._styled_mode = bool(value)

    @property
    def type(self) -> Optional[str]:
        """The default series type for the chart. Defaults to ``'line'``.

        Can be any of the chart types listed under :class:`PlotOptions` and
        :class:`Series`, or can be a series provided by an additional module.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type

    @type.setter
    def type(self, value):
        self._type = validators.string(value, allow_empty = True)

    @property
    def width(self) -> Optional[constants.EnforcedNullType | int | float | Decimal | str]:
        """An explicit width for the chart.

        Defaults to :class:`EnforcedNull <EnforcedNullType>`` which indicates
        a JavaScript value of ``null`` (as opposed to :obj:`None <python:None>` which
        results in ``undefined`` when converted JavaScript.)

        If a number, the width is given in pixels. If given a percentage string (for
        example ``'56%'``), the width is given as the percentage of the actual chart
        width. This allows for preserving the aspect ratio across responsive sizes.

        By default (when null) the width is calculated from the offset width of the
        containing element.

        :rtype: :class:`EnforcedNullType` or numeric or :class:`str <python:str>` or
          :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        if value is None:
            self._width = None
        else:
            try:
                try:
                    self._width = validators.numeric(value,
                                                      allow_empty = False,
                                                      minimum = 0)
                except ValueError:
                    self._width = validators.string(value,
                                                     allow_empty = False)
            except ValueError:
                raise errors.HighchartsValueError('Unable to resolve value to a '
                                                  'supported data type.')

    @property
    def zooming(self) -> Optional[ZoomingOptions]:
        """Chart zooming configuration.

        :rtype: :class:`ZoomingOptions <highcharts_maps.options.chart.zooming.ZoomingOptions>`
          or :obj:`None <python:None>`
        """
        return self._zooming

    @zooming.setter
    @class_sensitive(ZoomingOptions)
    def zooming(self, value):
        self._zooming = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'align_thresholds': as_dict.get('alignThresholds', None),
            'align_ticks': as_dict.get('alignTicks', None),
            'allow_mutating_data': as_dict.get('allowMutatingData', None),
            'animation': as_dict.get('animation', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'color_count': as_dict.get('colorCount', None),
            'display_errors': as_dict.get('displayErrors', None),
            'events': as_dict.get('events', None),
            'height': as_dict.get('height', None),
            'ignore_hidden_series': as_dict.get('ignoreHiddenSeries', None),
            'inverted': as_dict.get('inverted', None),
            'margin': as_dict.get('margin', None),
            'margin_bottom': as_dict.get('marginBottom', None),
            'margin_left': as_dict.get('marginLeft', None),
            'margin_right': as_dict.get('marginRight', None),
            'margin_top': as_dict.get('marginTop', None),
            'number_formatter': as_dict.get('numberFormatter', None),
            'options_3d': as_dict.get('options3d', None),
            'pan_key': as_dict.get('panKey', None),
            'panning': as_dict.get('panning', None),
            'parallel_axes': as_dict.get('parallelAxes', None),
            'parallel_coordinates': as_dict.get('parallelCoordinates', None),
            'plot_background_color': as_dict.get('plotBackgroundColor', None),
            'plot_background_image': as_dict.get('plotBackgroundImage', None),
            'plot_border_color': as_dict.get('plotBorderColor', None),
            'plot_border_width': as_dict.get('plotBorderWidth', None),
            'plot_shadow': as_dict.get('plotShadow', None),
            'polar': as_dict.get('polar', None),
            'reflow': as_dict.get('reflow', None),
            'render_to': as_dict.get('renderTo', None),
            'scrollable_plot_area': as_dict.get('scrollablePlotArea', None),
            'selection_marker_fill': as_dict.get('selectionMarkerFill', None),
            'shadow': as_dict.get('shadow', None),
            'show_axes': as_dict.get('showAxes', None),
            'spacing': as_dict.get('spacing', None),
            'spacing_bottom': as_dict.get('spacingBottom', None),
            'spacing_left': as_dict.get('spacingLeft', None),
            'spacing_top': as_dict.get('spacingTop', None),
            'spacing_right': as_dict.get('spacingRight', None),
            'style': as_dict.get('style', None),
            'styled_mode': as_dict.get('styledMode', None),
            'type': as_dict.get('type', None),
            'width': as_dict.get('width', None),
            'zooming': as_dict.get('zooming', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'alignThresholds': self.align_thresholds,
            'alignTicks': self.align_ticks,
            'allowMutatingData': self.allow_mutating_data,
            'animation': self.animation,
            'backgroundColor': self.background_color,
            'borderColor': self.border_color,
            'borderRadius': self.border_radius,
            'borderWidth': self.border_width,
            'className': self.class_name,
            'colorCount': self.color_count,
            'displayErrors': self.display_errors,
            'events': self.events,
            'height': self.height,
            'ignoreHiddenSeries': self.ignore_hidden_series,
            'inverted': self.inverted,
            'margin': self.margin,
            'marginBottom': self.margin_bottom,
            'marginLeft': self.margin_left,
            'marginRight': self.margin_right,
            'marginTop': self.margin_top,
            'numberFormatter': self.number_formatter,
            'options3d': self.options_3d,
            'panKey': self.pan_key,
            'panning': self.panning,
            'parallelAxes': self.parallel_axes,
            'parallelCoordinates': self.parallel_coordinates,
            'plotBackgroundColor': self.plot_background_color,
            'plotBackgroundImage': self.plot_background_image,
            'plotBorderColor': self.plot_border_color,
            'plotBorderWidth': self.plot_border_width,
            'plotShadow': self.plot_shadow,
            'polar': self.polar,
            'reflow': self.reflow,
            'renderTo': self.render_to,
            'scrollablePlotArea': self.scrollable_plot_area,
            'selectionMarkerFill': self.selection_marker_fill,
            'shadow': self.shadow,
            'showAxes': self.show_axes,
            'spacing': self.spacing,
            'spacingBottom': self.spacing_bottom,
            'spacingLeft': self.spacing_left,
            'spacingRight': self.spacing_right,
            'spacingTop': self.spacing_top,
            'style': self.style,
            'styledMode': self.styled_mode,
            'type': self.type,
            'width': self.width,
            'zooming': self.zooming,
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import errors, utility_functions
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class PartialFillOptions(HighchartsMeta):
    """A partial fill for each point, typically used to visualize how much of a task is
    performed.

    .. note::

      The partial fill object can be set either on series or point level.

    """

    def __init__(self, **kwargs):
        self._fill = None

        self.fill = kwargs.get('fill', None)

    @property
    def fill(self) -> Optional[str | Gradient | Pattern]:
        """The fill color to be used for partial fills. Defaults to
        :obj:`None <python:None>`, which applies a darker shade of the point color.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._fill

    @fill.setter
    def fill(self, value):
        self._fill = utility_functions.validate_color(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        return {
            'fill': as_dict.get('fill', None)
        }

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'fill': self.fill
        }



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern

from highcharts_core.options.axes.generic import GenericAxis
from highcharts_core.options.axes.data_classes import DataClass
from highcharts_core.options.axes.markers import AxisMarker


class ColorAxis(GenericAxis):
    """A color axis for series.

    Visually, the color axis will appear as a gradient or as separate items inside the
    legend, depending on whether the axis is scalar or based on data classes.

    .. seealso::

      For supported color formats, see the
      `documentation article about colors <https://www.highcharts.com/docs/chart-design-and-style/colors>`_.

    A scalar color axis is represented by a gradient. The colors either range between the
    :meth:`ColorAxis.min_color` and the :meth:`ColorAxis.max_color`, or for more
    fine-grained control the colors can be defined in :meth:`ColorAxis.stops`. Often
    times, the color axis needs to be adjusted to get the right color spread for the data.
    In addition to :meth:`stops <ColorAxis.stops>`, consider using a logarithmic axis
    type, or setting min and max to avoid the colors being determined by outliers.

    .. note::

      When :meth:`ColorAxis.data_classes` are used, the ranges are subdivided into
      separate classes like categories based on their values. This can be used for ranges
      between two values, but also for a true category. However, when your data is
      categorized, it may be as convenient to add each category to a separate series.

    .. warning::

      :class:`ColorAxis` does not work with the following series types:

        * :term:`sankey <Sankey Chart>`
        * :term:`sunburst <Sunburst>`
        * :term:`dependency wheels <Dependency Wheel>`
        * :term:`network graphs <Network Graph>`
        * :term:`wordclouds <Wordcloud>`
        * :term:`venn diagrams <Venn Diagram>`
        * :term:`solidgauge <SolidGauge>`

    """

    def __init__(self, **kwargs):
        self._data_class_color = None
        self._data_classes = None
        self._layout = None
        self._line_color = None
        self._marker = None
        self._max_color = None
        self._min_color = None
        self._show_in_legend = None
        self._stops = None

        self.data_class_color = kwargs.get('data_class_color', None)
        self.data_classes = kwargs.get('data_classes', None)
        self.layout = kwargs.get('layout', None)
        self.line_color = kwargs.get('line_color', None)
        self.marker = kwargs.get('marker', None)
        self.max_color = kwargs.get('max_color', None)
        self.min_color = kwargs.get('min_color', None)
        self.show_in_legend = kwargs.get('show_in_legend', None)
        self.stops = kwargs.get('stops', None)

        super().__init__(**kwargs)

    @property
    def data_class_color(self) -> Optional[str]:
        """Determines how to set each data class' color if no individual color is set.
        The default value, ``'tween'``, computes intermediate colors between
        :meth:`ColorAxis.min_color` and :meth:`ColorAxis.max_color`. The other possible
        value, ``'category'``, pulls colors from the global or chart specific ``colors``
        array.

        Defaults to ``'tween'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._data_class_color

    @data_class_color.setter
    def data_class_color(self, value):
        if not value:
            self._data_class_color = None
        else:
            value = validators.string(value)
            if value not in ['tween', 'category']:
                raise errors.HighchartsValueError(f'data_class_color expects either '
                                                  f'"tween" or "category". Received: '
                                                  f'{value}')
            self._data_class_color = value

    @property
    def data_classes(self) -> Optional[List[DataClass]]:
        """A collection of data classes or ranges for the choropleth map. If
        :obj:`None <python:None>`, the color axis will be scalar and values will be
        distributed as a gradient between the minimum and maximum colors. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`list <python:list>` of :class:`DataClass` instances, or
          :obj:`None <python:None>`
        """
        return self._data_classes

    @data_classes.setter
    @class_sensitive(DataClass, force_iterable = True)
    def data_classes(self, value):
        self._data_classes = value

    @property
    def layout(self) -> Optional[str]:
        """The layout of the color axis. Defaults to :obj:`None <python:None>`.

        Accepts either:

          * ``'horizontal'``
          * ``'vertical'``

        If :obj:`None <python:None>`, the color axis will have the same layout as the
        legend.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._layout

    @layout.setter
    def layout(self, value):
        if not value:
            self._layout = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['horizontal', 'vertical']:
                raise errors.HighchartsValueError(f'layout expects either '
                                                  f'"horizontal" or "vertical", but '
                                                  f'received: {value}')

            self._layout = value

    @property
    def line_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the line marking the axis itself. Defaults to ``'#ccd6eb'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        from highcharts_core import utility_functions
        self._line_color = utility_functions.validate_color(value)

    @property
    def marker(self) -> Optional[AxisMarker]:
        """The triangular marker on a scalar color axis that points to the value of the
        hovered area. To disable the marker, set the value to :obj:`None <python:None>`.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`AxisMarker` or :obj:`None <python:None>`
        """
        return self._marker

    @marker.setter
    @class_sensitive(AxisMarker)
    def marker(self, value):
        self._marker = value

    @property
    def max_color(self) -> Optional[str | Gradient | Pattern]:
        """The color used to represent the maximum value of the color axis. Defaults to
        ``'#003399'``.

        Unless :meth:`data_classes <ColorAxis.data_classes>` or
        :meth:`stops <ColorAxis.stops>` are set, the gradient ends at this value.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._max_color

    @max_color.setter
    def max_color(self, value):
        from highcharts_core import utility_functions
        self._max_color = utility_functions.validate_color(value)

    @property
    def min_color(self) -> Optional[str | Gradient | Pattern]:
        """The color used to represent the minimum value of the color axis. Defaults to
        ``'#e6ebf5'``.

        Unless :meth:`data_classes <ColorAxis.data_classes>` or
        :meth:`stops <ColorAxis.stops>` are set, the gradient starts at this value.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._min_color

    @min_color.setter
    def min_color(self, value):
        from highcharts_core import utility_functions
        self._min_color = utility_functions.validate_color(value)

    @property
    def show_in_legend(self) -> Optional[bool]:
        """Whether to display the color axis in the legend. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>`
        """
        return self._show_in_legend

    @show_in_legend.setter
    def show_in_legend(self, value):
        if value is None:
            self._show_in_legend = None
        else:
            self._show_in_legend = bool(value)

    @property
    def stops(self) -> Optional[List[List[int | float | Decimal | str]]]:
        """Color stops for the gradient in a color axis. Defaults to
        :obj:`None <python:None>`.

        .. hint::

          Use this in cases where a linear gradient between a
          :meth:`min_color <ColorAxis.min_color>` and
          :meth:`max_color <ColorAxis.max_color>` is not sufficient.

        .. note::

          Expects an iterable of 2-member iterables. Each 2-member iterable should be
          composed of a number and a string. The number should be a value between 0 and
          1 which assigns the relative position of the color in the gradient, while the
          string should represent the color itself.

        :rtype: :class:`list <python:list>` of :class:`list <python:list>`, where each
          second :class:`list <python:list>` consists of a :class:`float <python:float>`
          and a :class:`str <python:str>` / or :obj:`None <python:None>`
        """
        return self._stops

    @stops.setter
    def stops(self, value):
        if not value:
            self._stops = None
        else:
            value = validators.iterable(value)
            processed_items = []
            for item in value:
                item = validators.iterable(item)
                if len(item) != 2:
                    raise errors.HighchartsValueError(f'stops expects a list of 2-'
                                                      f'member iterables. Received a '
                                                      f'{len(item)}-member iterable.')
                item = [validators.float(item[0],
                                         minimum = 0,
                                         maximum = 1),
                        validators.string(item[1])]
                processed_items.append(item)

            self._stops = processed_items

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'angle': as_dict.get('angle', None),
            'ceiling': as_dict.get('ceiling', None),
            'class_name': as_dict.get('className', None),
            'end_on_tick': as_dict.get('endOnTick', None),
            'events': as_dict.get('events', None),
            'floor': as_dict.get('floor', None),
            'grid_line_color': as_dict.get('gridLineColor', None),
            'grid_line_dash_style': as_dict.get('gridLineDashStyle', None),
            'grid_line_interpolation': as_dict.get('gridLineInterpolation', None),
            'grid_line_width': as_dict.get('gridLineWidth', None),
            'grid_z_index': as_dict.get('gridZIndex', None),
            'id': as_dict.get('id', None),
            'labels': as_dict.get('labels', None),
            'margin': as_dict.get('margin', None),
            'max': as_dict.get('max', None),
            'max_padding': as_dict.get('maxPadding', None),
            'min': as_dict.get('min', None),
            'minor_grid_line_color': as_dict.get('minorGridLineColor', None),
            'minor_grid_line_dash_style': as_dict.get('minorGridLineDashStyle', None),
            'minor_grid_line_width': as_dict.get('minorGridLineWidth', None),
            'minor_tick_color': as_dict.get('minorTickColor', None),
            'minor_tick_interval': as_dict.get('minorTickInterval', None),
            'minor_tick_length': as_dict.get('minorTickLength', None),
            'minor_tick_position': as_dict.get('minorTickPosition', None),
            'minor_ticks': as_dict.get('minorTicks', None),
            'minor_tick_width': as_dict.get('minorTickWidth', None),
            'min_padding': as_dict.get('minPadding', None),
            'panning_enabled': as_dict.get('panningEnabled', None),
            'reversed': as_dict.get('reversed', None),
            'show_first_label': as_dict.get('showFirstLabel', None),
            'show_last_label': as_dict.get('showLastLabel', None),
            'soft_max': as_dict.get('softMax', None),
            'soft_min': as_dict.get('softMin', None),
            'start_of_week': as_dict.get('startOfWeek', None),
            'start_on_tick': as_dict.get('startOnTick', None),
            'tick_amount': as_dict.get('tickAmount', None),
            'tick_color': as_dict.get('tickColor', None),
            'tick_interval': as_dict.get('tickInterval', None),
            'tick_length': as_dict.get('tickLength', None),
            'tickmark_placement': as_dict.get('tickmarkPlacement', None),
            'tick_pixel_interval': as_dict.get('tickPixelInterval', None),
            'tick_position': as_dict.get('tickPosition', None),
            'tick_positioner': as_dict.get('tickPositioner', None),
            'tick_positions': as_dict.get('tickPositions', None),
            'tick_width': as_dict.get('tickWidth', None),
            'type': as_dict.get('type', None),
            'unique_names': as_dict.get('uniqueNames', None),
            'units': as_dict.get('units', None),
            'visible': as_dict.get('visible', None),
            'z_index': as_dict.get('zIndex', None),

            'data_class_color': as_dict.get('dataClassColor', None),
            'data_classes': as_dict.get('dataClasses', None),
            'layout': as_dict.get('layout', None),
            'line_color': as_dict.get('lineColor', None),
            'marker': as_dict.get('marker', None),
            'max_color': as_dict.get('maxColor', None),
            'min_color': as_dict.get('minColor', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'stops': as_dict.get('stops', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'angle': self.angle,
            'ceiling': self.ceiling,
            'className': self.class_name,
            'endOnTick': self.end_on_tick,
            'events': self.events,
            'floor': self.floor,
            'gridLineColor': self.grid_line_color,
            'gridLineDashStyle': self.grid_line_dash_style,
            'gridLineInterpolation': self.grid_line_interpolation,
            'gridLineWidth': self.grid_line_width,
            'gridZIndex': self.grid_z_index,
            'id': self.id,
            'labels': self.labels,
            'margin': self.margin,
            'max': self.max,
            'maxPadding': self.max_padding,
            'min': self.min,
            'minorGridLineColor': self.minor_grid_line_color,
            'minorGridLineDashStyle': self.minor_grid_line_dash_style,
            'minorGridLineWidth': self.minor_grid_line_width,
            'minorTickColor': self.minor_tick_color,
            'minorTickInterval': self.minor_tick_interval,
            'minorTickLength': self.minor_tick_length,
            'minorTickPosition': self.minor_tick_position,
            'minorTicks': self.minor_ticks,
            'minorTickWidth': self.minor_tick_width,
            'minPadding': self.min_padding,
            'panningEnabled': self.panning_enabled,
            'reversed': self.reversed,
            'showFirstLabel': self.show_first_label,
            'showLastLabel': self.show_last_label,
            'softMax': self.soft_max,
            'softMin': self.soft_min,
            'startOfWeek': self.start_of_week,
            'startOnTick': self.start_on_tick,
            'tickAmount': self.tick_amount,
            'tickColor': self.tick_color,
            'tickInterval': self.tick_interval,
            'tickLength': self.tick_length,
            'tickmarkPlacement': self.tickmark_placement,
            'tickPixelInterval': self.tick_pixel_interval,
            'tickPosition': self.tick_position,
            'tickPositioner': self.tick_positioner,
            'tickPositions': self.tick_positions,
            'tickWidth': self.tick_width,
            'type': self.type,
            'uniqueNames': self.unique_names,
            'units': self.units,
            'visible': self.visible,
            'zIndex': self.z_index,

            'dataClassColor': self.data_class_color,
            'dataClasses': self.data_classes,
            'layout': self.layout,
            'lineColor': self.line_color,
            'marker': self.marker,
            'maxColor': self.max_color,
            'minColor': self.min_color,
            'showInLegend': self.show_in_legend,
            'stops': self.stops
        }

        return untrimmed



from typing import Optional

import json

from highcharts_core import errors

from highcharts_core.options.series.base import SeriesBase
from highcharts_core.options.series.arcdiagram import ArcDiagramSeries
from highcharts_core.options.series.area import AreaSeries
from highcharts_core.options.series.area import AreaRangeSeries
from highcharts_core.options.series.area import AreaSplineSeries
from highcharts_core.options.series.area import AreaSplineRangeSeries
from highcharts_core.options.series.area import LineSeries
from highcharts_core.options.series.area import StreamGraphSeries
from highcharts_core.options.series.bar import BarSeries
from highcharts_core.options.series.bar import ColumnSeries
from highcharts_core.options.series.bar import ColumnPyramidSeries
from highcharts_core.options.series.bar import ColumnRangeSeries
from highcharts_core.options.series.bar import CylinderSeries
from highcharts_core.options.series.bar import VariwideSeries
from highcharts_core.options.series.bar import WaterfallSeries
from highcharts_core.options.series.bar import WindBarbSeries
from highcharts_core.options.series.bar import XRangeSeries
from highcharts_core.options.series.bellcurve import BellCurveSeries
from highcharts_core.options.series.boxplot import BoxPlotSeries
from highcharts_core.options.series.boxplot import ErrorBarSeries
from highcharts_core.options.series.bubble import BubbleSeries
from highcharts_core.options.series.bullet import BulletSeries
from highcharts_core.options.series.dependencywheel import DependencyWheelSeries
from highcharts_core.options.series.dumbbell import DumbbellSeries
from highcharts_core.options.series.dumbbell import LollipopSeries
from highcharts_core.options.series.funnel import FunnelSeries
from highcharts_core.options.series.funnel import Funnel3DSeries
from highcharts_core.options.series.gauge import GaugeSeries
from highcharts_core.options.series.gauge import SolidGaugeSeries
from highcharts_core.options.series.heatmap import HeatmapSeries
from highcharts_core.options.series.heatmap import TilemapSeries
from highcharts_core.options.series.histogram import HistogramSeries
from highcharts_core.options.series.item import ItemSeries
from highcharts_core.options.series.networkgraph import NetworkGraphSeries
from highcharts_core.options.series.organization import OrganizationSeries
from highcharts_core.options.series.packedbubble import PackedBubbleSeries
from highcharts_core.options.series.pareto import ParetoSeries
from highcharts_core.options.series.pictorial import PictorialSeries
from highcharts_core.options.series.pie import PieSeries
from highcharts_core.options.series.pie import VariablePieSeries
from highcharts_core.options.series.polygon import PolygonSeries
from highcharts_core.options.series.pyramid import PyramidSeries
from highcharts_core.options.series.pyramid import Pyramid3DSeries
from highcharts_core.options.series.sankey import SankeySeries
from highcharts_core.options.series.scatter import ScatterSeries
from highcharts_core.options.series.scatter import Scatter3DSeries
from highcharts_core.options.series.spline import SplineSeries
from highcharts_core.options.series.sunburst import SunburstSeries
from highcharts_core.options.series.timeline import TimelineSeries
from highcharts_core.options.series.treegraph import TreegraphSeries
from highcharts_core.options.series.treemap import TreemapSeries
from highcharts_core.options.series.vector import VectorSeries
from highcharts_core.options.series.venn import VennSeries
from highcharts_core.options.series.wordcloud import WordcloudSeries


SERIES_CLASSES = {
    'arcdiagram': ArcDiagramSeries,
    'area': AreaSeries,
    'arearange': AreaRangeSeries,
    'areaspline': AreaSplineSeries,
    'areasplinerange': AreaSplineRangeSeries,
    'line': LineSeries,
    'streamgraph': StreamGraphSeries,
    'bar': BarSeries,
    'column': ColumnSeries,
    'columnpyramid': ColumnPyramidSeries,
    'columnrange': ColumnRangeSeries,
    'cylinder': CylinderSeries,
    'variwide': VariwideSeries,
    'waterfall': WaterfallSeries,
    'windbarb': WindBarbSeries,
    'xrange': XRangeSeries,
    'bellcurve': BellCurveSeries,
    'boxplot': BoxPlotSeries,
    'errorbar': ErrorBarSeries,
    'bubble': BubbleSeries,
    'bullet': BulletSeries,
    'dependencywheel': DependencyWheelSeries,
    'dumbbell': DumbbellSeries,
    'lollipop': LollipopSeries,
    'funnel': FunnelSeries,
    'funnel3d': Funnel3DSeries,
    'gauge': GaugeSeries,
    'solidgauge': SolidGaugeSeries,
    'heatmap': HeatmapSeries,
    'tilemap': TilemapSeries,
    'histogram': HistogramSeries,
    'item': ItemSeries,
    'networkgraph': NetworkGraphSeries,
    'organization': OrganizationSeries,
    'packedbubble': PackedBubbleSeries,
    'pareto': ParetoSeries,
    'pictorial': PictorialSeries,
    'pie': PieSeries,
    'variablepie': VariablePieSeries,
    'polygon': PolygonSeries,
    'pyramid': PyramidSeries,
    'pyramid3d': Pyramid3DSeries,
    'sankey': SankeySeries,
    'scatter': ScatterSeries,
    'scatter3d': Scatter3DSeries,
    'spline': SplineSeries,
    'sunburst': SunburstSeries,
    'timeline': TimelineSeries,
    'treegraph': TreegraphSeries,
    'treemap': TreemapSeries,
    'vector': VectorSeries,
    'venn': VennSeries,
    'wordcloud': WordcloudSeries,
}


def create_series_obj(value, default_type = None) -> Optional[SeriesBase]:
    """Create an instance descended from :class:`SeriesBase`.

    :param value: The input that should be de-serialized to a :class:`SeriesBase`
      instance. Expected to be either a :class:`SeriesBase` instance, a
      :class:`dict <python:dict>`, or a JSON :class:`str <python:str>`

    :param default_type: The default series type to apply if not specified in
      ``value``. Defaults to :obj:`None <python:None>`
    :type default_type: :class:`str <python:str>` or :obj:`None <python:None>`

    :returns: A :class:`SeriesBase` (descendant) instance
    :rtype: :class:`SeriesBase` or :obj:`None <python:None>`
    """
    if not value:
        return None

    if isinstance(value, SeriesBase):
        return value

    if isinstance(value, dict):
        type_ = value.get('type', default_type)
        if not type_:
            raise errors.HighchartsValueError('To instantiate a Series, the "type" must '
                                              'be provided. No "type" key was found.')
        cls = SERIES_CLASSES.get(type_, default_type)
        if not cls:
            raise errors.HighchartsValueError(f'create_series_obj expects a value with '
                                              f'a recognized "type". However, received a '
                                              f'"type" value that was not recognized: '
                                              f'{type_}')

        instance = cls.from_dict(value)
    elif isinstance(value, str):
        try:
            preliminary = SeriesBase.from_js_literal(value)
            type_ = preliminary.type
        except errors.HighchartsParseError:
            preliminary_as_dict = json.loads(value)

            type_ = preliminary_as_dict.get('type', default_type)

        if not type_:
            raise errors.HighchartsValueError('To instantiate a Series, the "type" must '
                                              'be provided. No "type" key was found.')
        cls = SERIES_CLASSES.get(type_, default_type)
        if not cls:
            raise errors.HighchartsValueError(f'create_series_obj expects a value with '
                                              f'a recognized "type". However, received a '
                                              f'"type" value that was not recognized: '
                                              f'{type_}')

        try:
            instance = cls.from_js_literal(value)
        except errors.HighchartsParseError:
            instance = cls.from_json(value)
    else:
        raise errors.HighchartsValueError(f'create_series_obj expects a value whose '
                                          f'type is a SeriesBase or coercable to it, '
                                          f'but received: {value.__class__.__name__}')

    return instance



"""Tests for ``highcharts.subtitle``."""

import pytest

from highcharts_core.options.subtitle import Subtitle as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
        'text': 'Subtitle aligned left',
        'align': 'left',
        'x': 70
    }, None),
    ({
        'text': 'Subtitle floating left',
        'floating': True,
        'align': 'left',
        'x': 100,
        'y': 50
    }, None),
    ({
        'use_html': True
    }, None),
    ({
        'vertical_align': 'middle'
    }, None),
    ({
        'width_adjust': -44
    }, None),

    ({'align': 'not-valid'}, errors.HighchartsValueError),
    ({'x': 'not-valid'}, TypeError),
    ({'vertical_align': 'not-valid'}, errors.HighchartsValueError),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('subtitle/01.js', False, None),
    ('subtitle/02.js', False, None),

    ('subtitle/error-01.js', False, errors.HighchartsValueError),
    ('subtitle/error-02.js', False, errors.HighchartsParseError),

    ('subtitle/01.js', True, None),
    ('subtitle/02.js', True, None),

    ('subtitle/error-01.js', True, errors.HighchartsValueError),
    ('subtitle/error-02.js', True, errors.HighchartsParseError),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class ZoomLanguageOptions(HighchartsMeta):
    """Chart and map zoom accessibility language options."""

    def __init__(self, **kwargs):
        self._map_zoom_in = None
        self._map_zoom_out = None
        self._reset_zoom_button = None

        self.map_zoom_in = kwargs.get('map_zoom_in', None)
        self.map_zoom_out = kwargs.get('map_zoom_out', None)
        self.reset_zoom_button = kwargs.get('reset_zoom_button', None)

    @property
    def map_zoom_in(self) -> Optional[str]:
        """Defaults to
        ``'Zoom chart'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map_zoom_in

    @map_zoom_in.setter
    def map_zoom_in(self, value):
        self._map_zoom_in = validators.string(value, allow_empty = True)

    @property
    def map_zoom_out(self) -> Optional[str]:
        """Defaults to
        ``'Zoom out chart'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._map_zoom_out

    @map_zoom_out.setter
    def map_zoom_out(self, value):
        self._map_zoom_out = validators.string(value, allow_empty = True)

    @property
    def reset_zoom_button(self) -> Optional[str]:
        """Defaults to
        ``'Reset zoom'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._reset_zoom_button

    @reset_zoom_button.setter
    def reset_zoom_button(self, value):
        self._reset_zoom_button = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'map_zoom_in': as_dict.get('mapZoomIn', None),
            'map_zoom_out': as_dict.get('mapZoomOut', None),
            'reset_zoom_button': as_dict.get('resetZoomButton', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'mapZoomIn': self.map_zoom_in,
            'mapZoomOut': self.map_zoom_out,
            'resetZoomButton': self.reset_zoom_button
        }

        return untrimmed
